'use strict';

var xstate = require('xstate');

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

const STATE_INTERRUPTED_ID = 'INTERRUPTED';
const STATE_INTERRUPTED_TARGET = `#${STATE_INTERRUPTED_ID}`;
const STATE_FINISHED_ID = 'FINISHED';
const STATE_FINISHED_TARGET = `#${STATE_FINISHED_ID}`;
const STATE_FAILED_ID = 'FAILED';
const STATE_FAILED_TARGET = `#${STATE_FAILED_ID}`;

class ActivityStateProvider {
    constructor(step, init) {
        this.step = step;
        this.init = init;
    }
    get(context, nodeId) {
        let activityState = context.activityStates[nodeId];
        if (!activityState) {
            activityState = this.init(this.step, context.globalState);
            context.activityStates[nodeId] = activityState;
        }
        return activityState;
    }
}

class MachineUnhandledError extends Error {
    constructor(message, cause, stepId) {
        super(message);
        this.cause = cause;
        this.stepId = stepId;
    }
}

const STATE_STEP_ID_PREFIX = 'STEP_';
const STATE_BRANCH_ID_PREFIX = 'BRANCH_';
function getStepNodeId(stepId) {
    return STATE_STEP_ID_PREFIX + stepId;
}
function getBranchNodeId(stepId, branchName) {
    return STATE_BRANCH_ID_PREFIX + stepId + '_' + branchName;
}

function readMetaPath(meta) {
    const path = meta.src.type;
    let start = path.lastIndexOf(STATE_STEP_ID_PREFIX);
    if (start < 0) {
        return null;
    }
    start += STATE_STEP_ID_PREFIX.length;
    let end = path.indexOf('.', start);
    if (end < 0) {
        end = path.indexOf(':', start);
    }
    return end < 0 ? path.substring(start) : path.substring(start, end);
}

function catchUnhandledError(callback) {
    return (context, event, meta) => __awaiter(this, void 0, void 0, function* () {
        try {
            yield callback(context, event, meta);
        }
        catch (e) {
            const message = e instanceof Error ? e.message : String(e);
            const stepId = readMetaPath(meta);
            context.unhandledError = new MachineUnhandledError(message, e, stepId);
            throw e;
        }
    });
}

function isInterruptResult(result) {
    return typeof result === 'object' && result.interrupt === true;
}
function interrupt() {
    return { interrupt: true };
}

class AtomActivityNodeBuilder {
    constructor(config) {
        this.config = config;
    }
    build(step, nextNodeTarget) {
        const activityStateProvider = new ActivityStateProvider(step, this.config.init);
        const nodeId = getStepNodeId(step.id);
        return {
            id: nodeId,
            invoke: {
                src: catchUnhandledError((context) => __awaiter(this, void 0, void 0, function* () {
                    const activityState = activityStateProvider.get(context, nodeId);
                    const result = yield this.config.handler(step, context.globalState, activityState);
                    if (isInterruptResult(result)) {
                        context.interrupted = nodeId;
                        return;
                    }
                })),
                onDone: [
                    {
                        target: STATE_INTERRUPTED_TARGET,
                        cond: (context) => Boolean(context.interrupted)
                    },
                    {
                        target: nextNodeTarget
                    }
                ],
                onError: STATE_FAILED_TARGET
            }
        };
    }
}

function createAtomActivity(stepType, config) {
    return {
        stepType: stepType,
        nodeBuilderFactory: () => new AtomActivityNodeBuilder(config)
    };
}
function createAtomActivityFromHandler(stepType, handler) {
    return createAtomActivity(stepType, {
        init: () => ({}),
        handler
    });
}

const stackKey = 'loopStack';
function getLoopStack(buildingContext) {
    let stack = buildingContext[stackKey];
    if (!stack) {
        stack = new LoopStack();
        buildingContext[stackKey] = stack;
    }
    return stack;
}
class LoopStack {
    constructor() {
        this.stack = [];
    }
    push(name, leaveNodeTarget) {
        if (this.stack.some(item => item.name === name)) {
            throw new Error(`Loop with name "${name}" is already used in the stack`);
        }
        this.stack.push({
            name,
            leaveNodeTarget
        });
    }
    pop() {
        this.stack.pop();
    }
    getCurrentName() {
        const item = this.stack[this.stack.length - 1];
        return item ? item.name : null;
    }
    getNodeTarget(nameOrIndex) {
        let item;
        if (typeof nameOrIndex === 'number') {
            if (nameOrIndex !== -1) {
                throw new Error(`Index ${nameOrIndex} is not supported`);
            }
            item = this.stack[this.stack.length - 1];
            if (!item) {
                throw new Error('Cannot find any parent loop');
            }
        }
        else {
            item = this.stack.find(item => item.name === nameOrIndex);
            if (!item) {
                throw new Error(`Loop "${nameOrIndex}" not found`);
            }
        }
        return item.leaveNodeTarget;
    }
}

class ActivitySet {
    constructor(activities) {
        this.activities = activities;
    }
    get(stepType) {
        const provider = this.activities.get(stepType);
        if (!provider) {
            throw new Error(`Cannot find activity for step type: ${stepType}`);
        }
        return provider;
    }
    getTypes() {
        return Array.from(this.activities.keys());
    }
}
function createActivitySet(activities) {
    const map = new Map();
    for (const activity of activities) {
        if (map.has(activity.stepType)) {
            throw new Error(`Duplicate step type: ${activity.stepType}`);
        }
        map.set(activity.stepType, activity);
    }
    return new ActivitySet(map);
}

class ActivityNodeBuilderResolver {
    constructor(activitiesSet, sequenceNodeBuilder) {
        this.activitiesSet = activitiesSet;
        this.sequenceNodeBuilder = sequenceNodeBuilder;
        this.builders = new Map();
    }
    resolve(stepType) {
        let builder = this.builders.get(stepType);
        if (builder) {
            return builder;
        }
        const activity = this.activitiesSet.get(stepType);
        builder = activity.nodeBuilderFactory(this.sequenceNodeBuilder);
        this.builders.set(stepType, builder);
        return builder;
    }
}

class SequenceNodeBuilder {
    constructor(activitySet) {
        this.activitySet = activitySet;
        this.resolver = new ActivityNodeBuilderResolver(this.activitySet, this);
    }
    build(buildingContext, sequence, nextNodeTarget) {
        if (sequence.length === 0) {
            return {
                invoke: {
                    src: () => Promise.resolve(),
                    onDone: { target: nextNodeTarget }
                }
            };
        }
        const states = {};
        for (let index = 0; index < sequence.length; index++) {
            const step = sequence[index];
            const nodeId = getStepNodeId(step.id);
            const nextStep = sequence[index + 1];
            let itemNextTarget;
            if (nextStep) {
                const nextNodeId = getStepNodeId(nextStep.id);
                itemNextTarget = `#${nextNodeId}`;
            }
            else {
                itemNextTarget = nextNodeTarget;
            }
            const nodeBuilder = this.resolver.resolve(step.type);
            states[nodeId] = nodeBuilder.build(step, itemNextTarget, buildingContext);
        }
        const firstNodeId = Object.keys(states)[0];
        return {
            initial: firstNodeId,
            states
        };
    }
}

function branchName(branchName) {
    return { branchName };
}
function isBranchNameResult(result) {
    return typeof result === 'object' && Boolean(result.branchName);
}

function isSkipResult(result) {
    return typeof result === 'object' && result.skip === true;
}
function skip() {
    return { skip: true };
}

function isBreakResult(result) {
    return typeof result === 'object' && result.break;
}
function break_() {
    return { break: true };
}

class BreakActivityNodeBuilder {
    constructor(config) {
        this.config = config;
    }
    build(step, nextNodeTarget, buildingContext) {
        const activityStateProvider = new ActivityStateProvider(step, (s, g) => {
            return {
                activityState: this.config.init(s, g)
            };
        });
        const nodeId = getStepNodeId(step.id);
        const loopNameOrIndex = this.config.loopName(step);
        const leaveNodeTarget = getLoopStack(buildingContext).getNodeTarget(loopNameOrIndex);
        return {
            id: nodeId,
            invoke: {
                src: catchUnhandledError((context) => __awaiter(this, void 0, void 0, function* () {
                    const internalState = activityStateProvider.get(context, nodeId);
                    const result = yield this.config.handler(step, context.globalState, internalState.activityState);
                    if (isInterruptResult(result)) {
                        context.interrupted = nodeId;
                        return;
                    }
                    internalState.break = isBreakResult(result);
                })),
                onDone: [
                    {
                        target: STATE_INTERRUPTED_TARGET,
                        cond: (context) => Boolean(context.interrupted)
                    },
                    {
                        target: leaveNodeTarget,
                        cond: (context) => {
                            const internalState = activityStateProvider.get(context, nodeId);
                            return Boolean(internalState.break);
                        }
                    },
                    {
                        target: nextNodeTarget
                    }
                ],
                onError: STATE_FAILED_TARGET
            }
        };
    }
}

function createBreakActivity(stepType, config) {
    return {
        stepType,
        nodeBuilderFactory: () => new BreakActivityNodeBuilder(config)
    };
}

class ContainerActivityNodeBuilder {
    constructor(sequenceNodeBuilder, config) {
        this.sequenceNodeBuilder = sequenceNodeBuilder;
        this.config = config;
    }
    build(step, nextNodeTarget, buildingContext) {
        const activityStateProvider = new ActivityStateProvider(step, this.config.init);
        const nodeId = getStepNodeId(step.id);
        const enterNodeId = `ENTER.${nodeId}`;
        const leaveNodeId = `LEAVE.${nodeId}`;
        const leaveNodeTarget = `#${leaveNodeId}`;
        const createState = (id, handle, nextStateNodeTarget) => {
            return {
                id,
                invoke: {
                    src: catchUnhandledError((context) => __awaiter(this, void 0, void 0, function* () {
                        if (handle) {
                            const activityState = activityStateProvider.get(context, nodeId);
                            const result = yield handle(step, context.globalState, activityState);
                            if (isInterruptResult(result)) {
                                context.interrupted = nodeId;
                                return;
                            }
                        }
                    })),
                    onDone: [
                        {
                            target: STATE_INTERRUPTED_TARGET,
                            cond: (context) => Boolean(context.interrupted)
                        },
                        {
                            target: nextStateNodeTarget
                        }
                    ],
                    onError: STATE_FAILED_TARGET
                }
            };
        };
        return {
            id: nodeId,
            initial: 'ENTER',
            states: {
                ENTER: createState(enterNodeId, this.config.onEnter, 'SEQUENCE'),
                SEQUENCE: this.sequenceNodeBuilder.build(buildingContext, step.sequence, leaveNodeTarget),
                LEAVE: createState(leaveNodeId, this.config.onLeave, nextNodeTarget)
            }
        };
    }
}

function createContainerActivity(stepType, config) {
    return {
        stepType,
        nodeBuilderFactory: sequenceNodeBuilder => new ContainerActivityNodeBuilder(sequenceNodeBuilder, config)
    };
}

class ForkActivityNodeBuilder {
    constructor(sequenceNodeBuilder, config) {
        this.sequenceNodeBuilder = sequenceNodeBuilder;
        this.config = config;
    }
    build(step, nextNodeTarget, buildingContext) {
        const activityStateProvider = new ActivityStateProvider(step, (s, g) => {
            return {
                activityState: this.config.init(s, g)
            };
        });
        const nodeId = getStepNodeId(step.id);
        const branchNames = Object.keys(step.branches);
        const states = {};
        for (const branchName of branchNames) {
            const branchNodeId = getBranchNodeId(step.id, branchName);
            states[branchNodeId] = this.sequenceNodeBuilder.build(buildingContext, step.branches[branchName], nextNodeTarget);
        }
        const CONDITION = {
            invoke: {
                src: catchUnhandledError((context) => __awaiter(this, void 0, void 0, function* () {
                    const internalState = activityStateProvider.get(context, nodeId);
                    const result = yield this.config.handler(step, context.globalState, internalState.activityState);
                    if (isInterruptResult(result)) {
                        context.interrupted = nodeId;
                        return;
                    }
                    if (isBranchNameResult(result)) {
                        if (!branchNames.includes(result.branchName)) {
                            throw new Error(`Branch ${result.branchName} does not exist`);
                        }
                        internalState.targetBranchName = result.branchName;
                        internalState.skipped = undefined;
                        return;
                    }
                    if (isSkipResult(result)) {
                        internalState.targetBranchName = undefined;
                        internalState.skipped = true;
                        return;
                    }
                    throw new Error('Not supported result for conditional activity');
                })),
                onDone: [
                    {
                        target: STATE_INTERRUPTED_TARGET,
                        cond: (context) => Boolean(context.interrupted)
                    },
                    {
                        target: nextNodeTarget,
                        cond: (context) => {
                            const activityState = activityStateProvider.get(context, nodeId);
                            return Boolean(activityState.skipped);
                        }
                    },
                    ...branchNames.map(branchName => {
                        return {
                            target: getBranchNodeId(step.id, branchName),
                            cond: (context) => {
                                const activityState = activityStateProvider.get(context, nodeId);
                                return activityState.targetBranchName === branchName;
                            }
                        };
                    })
                ],
                onError: STATE_FAILED_TARGET
            }
        };
        return {
            id: nodeId,
            initial: 'CONDITION',
            states: Object.assign({ CONDITION }, states)
        };
    }
}

function createForkActivity(stepType, config) {
    return {
        stepType,
        nodeBuilderFactory: sequenceNodeBuilder => new ForkActivityNodeBuilder(sequenceNodeBuilder, config)
    };
}

class InterruptionActivityNodeBuilder {
    constructor(config) {
        this.config = config;
    }
    build(step) {
        const nodeId = getStepNodeId(step.id);
        return {
            id: nodeId,
            invoke: {
                src: catchUnhandledError((context) => this.config.handler(step, context.globalState)),
                onDone: [
                    {
                        target: STATE_INTERRUPTED_TARGET
                    }
                ],
                onError: STATE_FAILED_TARGET
            }
        };
    }
}

function createInterruptionActivity(stepType, config) {
    return {
        stepType,
        nodeBuilderFactory: () => new InterruptionActivityNodeBuilder(config)
    };
}

class LoopActivityNodeBuilder {
    constructor(sequenceNodeBuilder, config) {
        this.sequenceNodeBuilder = sequenceNodeBuilder;
        this.config = config;
    }
    build(step, nextNodeTarget, buildingContext) {
        const activityStateProvider = new ActivityStateProvider(step, (s, g) => ({
            activityState: this.config.init(s, g)
        }));
        const nodeId = getStepNodeId(step.id);
        const conditionNodeId = `CONDITION.${nodeId}`;
        const conditionNodeTarget = `#${conditionNodeId}`;
        const leaveNodeId = `LEAVE.${nodeId}`;
        const leaveNodeTarget = `#${leaveNodeId}`;
        const loopName = this.config.loopName(step);
        const loopStack = getLoopStack(buildingContext);
        loopStack.push(loopName, leaveNodeTarget);
        const LOOP = this.sequenceNodeBuilder.build(buildingContext, step.sequence, conditionNodeTarget);
        loopStack.pop();
        return {
            id: nodeId,
            initial: 'ENTER',
            states: {
                ENTER: {
                    invoke: {
                        src: catchUnhandledError((context) => __awaiter(this, void 0, void 0, function* () {
                            if (this.config.onEnter) {
                                const internalState = activityStateProvider.get(context, nodeId);
                                const promise = this.config.onEnter(step, context.globalState, internalState.activityState);
                                if (promise) {
                                    yield promise;
                                }
                            }
                        })),
                        onDone: 'CONDITION',
                        onError: STATE_FAILED_TARGET
                    }
                },
                CONDITION: {
                    id: conditionNodeId,
                    invoke: {
                        src: catchUnhandledError((context) => __awaiter(this, void 0, void 0, function* () {
                            const internalState = activityStateProvider.get(context, nodeId);
                            const result = yield this.config.condition(step, context.globalState, internalState.activityState);
                            if (isInterruptResult(result)) {
                                context.interrupted = nodeId;
                                return;
                            }
                            internalState.continue = result;
                        })),
                        onDone: [
                            {
                                target: STATE_INTERRUPTED_TARGET,
                                cond: (context) => Boolean(context.interrupted)
                            },
                            {
                                target: 'LOOP',
                                cond: (context) => {
                                    const activityState = activityStateProvider.get(context, nodeId);
                                    return Boolean(activityState.continue);
                                }
                            },
                            {
                                target: 'LEAVE'
                            }
                        ],
                        onError: STATE_FAILED_TARGET
                    }
                },
                LOOP,
                LEAVE: {
                    id: leaveNodeId,
                    invoke: {
                        src: catchUnhandledError((context) => __awaiter(this, void 0, void 0, function* () {
                            if (this.config.onLeave) {
                                const internalState = activityStateProvider.get(context, nodeId);
                                const promise = this.config.onLeave(step, context.globalState, internalState.activityState);
                                if (promise) {
                                    yield promise;
                                }
                            }
                        })),
                        onDone: nextNodeTarget,
                        onError: STATE_FAILED_TARGET
                    }
                }
            }
        };
    }
}

function createLoopActivity(stepType, config) {
    return {
        stepType,
        nodeBuilderFactory: sequenceNodeBuilder => new LoopActivityNodeBuilder(sequenceNodeBuilder, config)
    };
}

class ParallelActivityNodeBuilder {
    constructor(sequenceNodeBuilder, config) {
        this.sequenceNodeBuilder = sequenceNodeBuilder;
        this.config = config;
    }
    build(step, nextNodeTarget, buildingContext) {
        const activityStateProvider = new ActivityStateProvider(step, (s, g) => ({ activityState: this.config.init(s, g) }));
        const nodeId = getStepNodeId(step.id);
        const branchNames = Object.keys(step.branches);
        const states = {};
        for (const branchName of branchNames) {
            const branchNodeId = getBranchNodeId(step.id, branchName);
            const leaveNodeId = branchNodeId + '_LEAVE';
            const leaveNodeTarget = '#' + leaveNodeId;
            states[branchNodeId] = {
                initial: 'TRY_LEAVE',
                states: {
                    TRY_LEAVE: {
                        invoke: {
                            src: catchUnhandledError(() => __awaiter(this, void 0, void 0, function* () {
                                // TODO: emit on enter?
                            })),
                            onDone: [
                                {
                                    target: 'SEQUENCE',
                                    cond: (context) => {
                                        var _a, _b;
                                        const activityState = activityStateProvider.get(context, nodeId);
                                        return (_b = (_a = activityState.branchNames) === null || _a === void 0 ? void 0 : _a.includes(branchName)) !== null && _b !== void 0 ? _b : false;
                                    }
                                },
                                {
                                    target: 'LEAVE'
                                }
                            ],
                            onError: STATE_FAILED_TARGET
                        }
                    },
                    SEQUENCE: this.sequenceNodeBuilder.build(buildingContext, step.branches[branchName], leaveNodeTarget),
                    LEAVE: {
                        id: leaveNodeId,
                        type: 'final'
                    }
                }
            };
        }
        const CONDITION = {
            invoke: {
                src: catchUnhandledError((context) => __awaiter(this, void 0, void 0, function* () {
                    const internalState = activityStateProvider.get(context, nodeId);
                    const result = yield this.config.handler(step, context.globalState, internalState.activityState);
                    if (isInterruptResult(result)) {
                        context.interrupted = nodeId;
                        return;
                    }
                    if (isSkipResult(result)) {
                        internalState.branchNames = [];
                        return;
                    }
                    if (Array.isArray(result)) {
                        internalState.branchNames = result.map(r => r.branchName);
                        return;
                    }
                    throw new Error('Not supported result for parallel activity');
                })),
                onDone: [
                    {
                        target: STATE_INTERRUPTED_TARGET,
                        cond: (context) => Boolean(context.interrupted)
                    },
                    {
                        target: 'PARALLEL'
                    }
                ],
                onError: STATE_FAILED_TARGET
            }
        };
        return {
            id: nodeId,
            initial: 'CONDITION',
            states: {
                CONDITION,
                PARALLEL: {
                    type: 'parallel',
                    states,
                    onDone: 'JOIN'
                },
                JOIN: {
                    invoke: {
                        src: catchUnhandledError((context) => __awaiter(this, void 0, void 0, function* () {
                            const internalState = activityStateProvider.get(context, nodeId);
                            internalState.branchNames = undefined;
                        })),
                        onDone: nextNodeTarget,
                        onError: STATE_FAILED_TARGET
                    }
                }
            }
        };
    }
}

function createParallelActivity(stepType, config) {
    return {
        stepType,
        nodeBuilderFactory: sequenceNodeBuilder => new ParallelActivityNodeBuilder(sequenceNodeBuilder, config)
    };
}

class SignalActivityNodeBuilder {
    constructor(config) {
        this.config = config;
    }
    build(step, nextNodeTarget) {
        const activityStateProvider = new ActivityStateProvider(step, this.config.init);
        const nodeId = getStepNodeId(step.id);
        return {
            id: nodeId,
            initial: 'BEFORE_SIGNAL',
            states: {
                BEFORE_SIGNAL: {
                    invoke: {
                        src: catchUnhandledError((context) => __awaiter(this, void 0, void 0, function* () {
                            const activityState = activityStateProvider.get(context, nodeId);
                            const result = yield this.config.beforeSignal(step, context.globalState, activityState);
                            if (isInterruptResult(result)) {
                                context.interrupted = nodeId;
                                return;
                            }
                        })),
                        onDone: [
                            {
                                target: STATE_INTERRUPTED_TARGET,
                                cond: (context) => Boolean(context.interrupted)
                            },
                            {
                                target: 'WAIT_FOR_SIGNAL'
                            }
                        ],
                        onError: STATE_FAILED_TARGET
                    }
                },
                WAIT_FOR_SIGNAL: {
                    on: {
                        SIGNAL_RECEIVED: {
                            target: 'AFTER_SIGNAL'
                        }
                    }
                },
                AFTER_SIGNAL: {
                    invoke: {
                        src: catchUnhandledError((context, event) => __awaiter(this, void 0, void 0, function* () {
                            const activityState = activityStateProvider.get(context, nodeId);
                            const ev = event;
                            const result = yield this.config.afterSignal(step, context.globalState, activityState, ev.payload);
                            if (isInterruptResult(result)) {
                                context.interrupted = nodeId;
                                return;
                            }
                        })),
                        onDone: [
                            {
                                target: STATE_INTERRUPTED_TARGET,
                                cond: (context) => Boolean(context.interrupted)
                            },
                            {
                                target: nextNodeTarget
                            }
                        ],
                        onError: STATE_FAILED_TARGET
                    }
                }
            }
        };
    }
}

function createSignalActivity(stepType, config) {
    return {
        stepType,
        nodeBuilderFactory: () => new SignalActivityNodeBuilder(config)
    };
}
function signalSignalActivity(interpreter, payload) {
    interpreter.sendSignal('SIGNAL_RECEIVED', payload);
}

function readStatePaths(stateValue) {
    const result = [];
    processPath(result, [], stateValue);
    return result;
}
function processPath(result, path, stateValue) {
    if (typeof stateValue === 'string') {
        path.push(stateValue);
        result.push(path);
    }
    else if (typeof stateValue === 'object') {
        const keys = Object.keys(stateValue);
        for (let i = 0; i < keys.length; i++) {
            const key = keys[i];
            const childPath = i === key.length - 1 ? path : [...path, key];
            processPath(result, childPath, stateValue[key]);
        }
    }
    else {
        throw new Error('Invalid state value');
    }
}

class WorkflowMachineSnapshot {
    constructor(globalState, unhandledError, stateValue) {
        this.globalState = globalState;
        this.unhandledError = unhandledError;
        this.stateValue = stateValue;
    }
    /**
     * @returns The paths of all currently executing states.
     * @example `[ ['MAIN', 'STEP_1'], ['MAIN', 'STEP_2'] ]`
     */
    getStatePaths() {
        if (!this.statePaths) {
            this.statePaths = readStatePaths(this.stateValue);
        }
        return this.statePaths;
    }
    /**
     * @returns The path of the currently executing state.
     * If multiple states are currently executing, it returns `null`.
     * To get all paths of executing states, use the `getStatePaths` method.
     * @example `['MAIN', 'STEP_1']`
     * @example `null`
     */
    tryGetStatePath() {
        const paths = this.getStatePaths();
        return paths.length === 1 ? paths[0] : null;
    }
    /**
     * @returns The ID of the currently executing state.
     * If multiple states are currently executing, it returns `null`.
     */
    tryGetCurrentStepId() {
        const path = this.tryGetStatePath();
        return path ? tryExtractStepId(path) : null;
    }
    /**
     * @returns the list of ID of the currently executing steps.
     */
    getCurrentStepIds() {
        const ids = new Set();
        const paths = this.getStatePaths();
        for (const path of paths) {
            const id = tryExtractStepId(path);
            if (id) {
                ids.add(id);
            }
        }
        return [...ids];
    }
    /**
     * @returns `true` if the workflow machine is finished, otherwise `false`.
     */
    isFinished() {
        const path = this.tryGetStatePath();
        return path !== null && path[0] === STATE_FINISHED_ID;
    }
    /**
     * @returns `true` if the workflow has failed, otherwise `false`.
     */
    isFailed() {
        const path = this.tryGetStatePath();
        return path !== null && path[0] === STATE_FAILED_ID;
    }
    /**
     * @returns `true` if the workflow is interrupted, otherwise `false`.
     */
    isInterrupted() {
        const path = this.tryGetStatePath();
        return path !== null && path[0] === STATE_INTERRUPTED_ID;
    }
}
function tryExtractStepId(path) {
    for (let i = path.length - 1; i >= 0; i--) {
        const item = path[i];
        if (item.startsWith(STATE_STEP_ID_PREFIX)) {
            return item.substring(STATE_STEP_ID_PREFIX.length);
        }
    }
    return null;
}

class WorkflowMachineInterpreter {
    constructor(interpreter, initState) {
        this.interpreter = interpreter;
        this.initState = initState;
    }
    start() {
        this.interpreter.start(this.initState);
        return this;
    }
    getSnapshot() {
        const snapshot = this.interpreter.getSnapshot();
        return new WorkflowMachineSnapshot(snapshot.context.globalState, snapshot.context.unhandledError, snapshot.value);
    }
    serializeSnapshot() {
        return this.interpreter.getSnapshot().toJSON();
    }
    onDone(callback) {
        this.interpreter.onStop(callback);
        return this;
    }
    onChange(callback) {
        this.interpreter.onChange(callback);
        return this;
    }
    isRunning() {
        return this.interpreter.status === xstate.InterpreterStatus.Running;
    }
    tryStop() {
        if (this.isRunning()) {
            this.interpreter.stop();
            return true;
        }
        return false;
    }
    sendSignal(event, payload) {
        this.interpreter.send(event, {
            payload
        });
        return this;
    }
    getNative() {
        return this.interpreter;
    }
}

class WorkflowMachine {
    constructor(definition, machine) {
        this.definition = definition;
        this.machine = machine;
    }
    create(config) {
        return this.restore({
            globalState: config.init(this.definition),
            activityStates: {}
        });
    }
    restore(context) {
        const machine = this.machine.withContext(context);
        return new WorkflowMachineInterpreter(xstate.interpret(machine), undefined);
    }
    deserializeSnapshot(serializedSnapshot) {
        const initState = this.machine.resolveState(xstate.State.create(serializedSnapshot));
        return new WorkflowMachineInterpreter(xstate.interpret(this.machine), initState);
    }
    getNative() {
        return this.machine;
    }
}

class WorkflowMachineBuilder {
    constructor(sequenceNodeBuilder) {
        this.sequenceNodeBuilder = sequenceNodeBuilder;
    }
    build(definition, config) {
        const buildingContext = {};
        const initial = (config === null || config === void 0 ? void 0 : config.initialStepId) ? getStepNodeId(config.initialStepId) : 'MAIN';
        const machine = xstate.createMachine({
            initial,
            predictableActionArguments: true,
            states: {
                MAIN: this.sequenceNodeBuilder.build(buildingContext, definition.sequence, STATE_FINISHED_TARGET),
                FINISHED: {
                    id: STATE_FINISHED_ID,
                    type: 'final'
                },
                INTERRUPTED: {
                    id: STATE_INTERRUPTED_ID,
                    type: 'final'
                },
                FAILED: {
                    id: STATE_FAILED_ID,
                    type: 'final'
                }
            }
        });
        return new WorkflowMachine(definition, machine);
    }
}
function createWorkflowMachineBuilder(activitySet) {
    return new WorkflowMachineBuilder(new SequenceNodeBuilder(activitySet));
}

exports.ActivityNodeBuilderResolver = ActivityNodeBuilderResolver;
exports.ActivitySet = ActivitySet;
exports.ActivityStateProvider = ActivityStateProvider;
exports.MachineUnhandledError = MachineUnhandledError;
exports.STATE_BRANCH_ID_PREFIX = STATE_BRANCH_ID_PREFIX;
exports.STATE_FAILED_ID = STATE_FAILED_ID;
exports.STATE_FAILED_TARGET = STATE_FAILED_TARGET;
exports.STATE_FINISHED_ID = STATE_FINISHED_ID;
exports.STATE_FINISHED_TARGET = STATE_FINISHED_TARGET;
exports.STATE_INTERRUPTED_ID = STATE_INTERRUPTED_ID;
exports.STATE_INTERRUPTED_TARGET = STATE_INTERRUPTED_TARGET;
exports.STATE_STEP_ID_PREFIX = STATE_STEP_ID_PREFIX;
exports.SequenceNodeBuilder = SequenceNodeBuilder;
exports.WorkflowMachine = WorkflowMachine;
exports.WorkflowMachineBuilder = WorkflowMachineBuilder;
exports.WorkflowMachineInterpreter = WorkflowMachineInterpreter;
exports.branchName = branchName;
exports.break_ = break_;
exports.catchUnhandledError = catchUnhandledError;
exports.createActivitySet = createActivitySet;
exports.createAtomActivity = createAtomActivity;
exports.createAtomActivityFromHandler = createAtomActivityFromHandler;
exports.createBreakActivity = createBreakActivity;
exports.createContainerActivity = createContainerActivity;
exports.createForkActivity = createForkActivity;
exports.createInterruptionActivity = createInterruptionActivity;
exports.createLoopActivity = createLoopActivity;
exports.createParallelActivity = createParallelActivity;
exports.createSignalActivity = createSignalActivity;
exports.createWorkflowMachineBuilder = createWorkflowMachineBuilder;
exports.getBranchNodeId = getBranchNodeId;
exports.getStepNodeId = getStepNodeId;
exports.interrupt = interrupt;
exports.isBranchNameResult = isBranchNameResult;
exports.isBreakResult = isBreakResult;
exports.isInterruptResult = isInterruptResult;
exports.isSkipResult = isSkipResult;
exports.signalSignalActivity = signalSignalActivity;
exports.skip = skip;
