import { StateNodeConfig, EventObject, StateMachine, StateSchema, Interpreter, Typestate, StateConfig, StateValue, State, InvokeMeta } from 'xstate';
import { Sequence, Step, Definition, SequentialStep, BranchedStep } from 'sequential-workflow-model';

declare class ActivitySet<TGlobalState> {
    private readonly activities;
    constructor(activities: Map<string, Activity<TGlobalState>>);
    get(stepType: string): Activity<TGlobalState>;
    getTypes(): string[];
}
declare function createActivitySet<TGlobalState>(activities: Activity<TGlobalState>[]): ActivitySet<TGlobalState>;

declare class SequenceNodeBuilder<TGlobalState> {
    private readonly activitySet;
    private readonly resolver;
    constructor(activitySet: ActivitySet<TGlobalState>);
    build(buildingContext: BuildingContext, sequence: Sequence, nextNodeTarget: string): ActivityNodeConfig<TGlobalState>;
}

declare class MachineUnhandledError extends Error {
    readonly cause: unknown;
    readonly stepId: string | null;
    constructor(message: string, cause: unknown, stepId: string | null);
}

declare const STATE_INTERRUPTED_ID = "INTERRUPTED";
declare const STATE_INTERRUPTED_TARGET: string;
declare const STATE_FINISHED_ID = "FINISHED";
declare const STATE_FINISHED_TARGET: string;
declare const STATE_FAILED_ID = "FAILED";
declare const STATE_FAILED_TARGET: string;
interface MachineContext<TGlobalState> {
    interrupted?: string;
    unhandledError?: MachineUnhandledError;
    globalState: TGlobalState;
    activityStates: Record<string, unknown>;
}
interface Activity<TGlobalState> {
    stepType: string;
    nodeBuilderFactory: ActivityNodeBuilderFactory<TGlobalState>;
}
type ActivityNodeBuilderFactory<TGlobalState> = (sequenceNodeBuilder: SequenceNodeBuilder<TGlobalState>) => ActivityNodeBuilder<TGlobalState>;
interface BuildingContext {
    [name: string]: unknown;
}
interface ActivityNodeBuilder<TGlobalState> {
    build(step: Step, nextNodeTarget: string, buildingContext: BuildingContext): ActivityNodeConfig<TGlobalState>;
}
type ActivityNodeConfig<TGlobalState> = StateNodeConfig<MachineContext<TGlobalState>, Record<string, any>, EventObject>;
type GlobalStateInitializer<TGlobalState> = (definition: Definition) => TGlobalState;
type ActivityStateInitializer<TStep extends Step, TGlobalState, TActivityState extends object> = (step: TStep, globalState: TGlobalState) => TActivityState;
type SequentialStateMachine<TGlobalState> = StateMachine<MachineContext<TGlobalState>, StateSchema, EventObject>;
type SequentialStateMachineInterpreter<TGlobalState> = Interpreter<MachineContext<TGlobalState>, StateSchema, EventObject, Typestate<MachineContext<TGlobalState>>, any>;
type SignalPayload = Record<string, unknown>;
type SerializedWorkflowMachineSnapshot<TGlobalState> = StateConfig<MachineContext<TGlobalState>, EventObject>;

interface InterruptResult {
    interrupt: true;
}
declare function isInterruptResult(result: unknown): result is InterruptResult;
declare function interrupt(): InterruptResult;

type AtomActivityHandler<TStep extends Step, TGlobalState, TActivityState> = (step: TStep, globalState: TGlobalState, activityState: TActivityState) => Promise<AtomActivityHandlerResult>;
type AtomActivityHandlerResult = void | InterruptResult;
interface AtomActivityConfig<TStep extends Step, TGlobalState, TActivityState extends object> {
    init: ActivityStateInitializer<TStep, TGlobalState, TActivityState>;
    handler: AtomActivityHandler<TStep, TGlobalState, TActivityState>;
}

declare function createAtomActivity<TStep extends Step = Step, TGlobalState = object, TActivityState extends object = object>(stepType: TStep['type'], config: AtomActivityConfig<TStep, TGlobalState, TActivityState>): Activity<TGlobalState>;
declare function createAtomActivityFromHandler<TStep extends Step = Step, TGlobalState = object>(stepType: TStep['type'], handler: AtomActivityHandler<TStep, TGlobalState, Record<string, never>>): Activity<TGlobalState>;

type BranchNameResult = {
    branchName: string;
};
declare function branchName(branchName: string): BranchNameResult;
declare function isBranchNameResult(result: unknown): result is BranchNameResult;

interface SkipResult {
    skip: true;
}
declare function isSkipResult(result: unknown): result is SkipResult;
declare function skip(): SkipResult;

interface BreakResult {
    break: true;
}
declare function isBreakResult(result: unknown): result is BreakResult;
declare function break_(): BreakResult;

type BreakActivityHandler<TStep extends Step, TGlobalState, TActivityState> = (step: TStep, globalState: TGlobalState, activityState: TActivityState) => Promise<BreakActivityHandlerResult>;
type BreakActivityHandlerResult = void | InterruptResult | BreakResult;
interface BreakActivityConfig<TStep extends Step, TGlobalState, TActivityState extends object> {
    loopName: (step: TStep) => string | -1;
    init: ActivityStateInitializer<TStep, TGlobalState, TActivityState>;
    handler: BreakActivityHandler<TStep, TGlobalState, TActivityState>;
}
interface BreakActivityState<TActivityState> {
    break?: boolean;
    activityState: TActivityState;
}

declare function createBreakActivity<TStep extends Step = Step, GlobalState = object, ActivityState extends object = object>(stepType: TStep['type'], config: BreakActivityConfig<TStep, GlobalState, ActivityState>): Activity<GlobalState>;

type ContainerActivityHandler<TStep extends Step, TGlobalState, TActivityState> = (step: TStep, globalState: TGlobalState, activityState: TActivityState) => Promise<ContainerActivityHandlerResult>;
type ContainerActivityHandlerResult = void | InterruptResult;
interface ContainerActivityConfig<TStep extends Step, TGlobalState, TActivityState extends object> {
    init: ActivityStateInitializer<TStep, TGlobalState, TActivityState>;
    onEnter?: ContainerActivityHandler<TStep, TGlobalState, TActivityState>;
    onLeave?: ContainerActivityHandler<TStep, TGlobalState, TActivityState>;
}

declare function createContainerActivity<TStep extends SequentialStep = SequentialStep, TGlobalState = object, TActivityState extends object = object>(stepType: TStep['type'], config: ContainerActivityConfig<TStep, TGlobalState, TActivityState>): Activity<TGlobalState>;

type ForkActivityHandlerResult = InterruptResult | BranchNameResult | SkipResult;
type ForkActivityHandler<TStep extends BranchedStep, TGlobalState, TActivityState> = (step: TStep, globalState: TGlobalState, activityState: TActivityState) => Promise<ForkActivityHandlerResult>;
interface ForkActivityConfig<TStep extends BranchedStep, TGlobalState, TActivityState extends object> {
    init: ActivityStateInitializer<TStep, TGlobalState, TActivityState>;
    handler: ForkActivityHandler<TStep, TGlobalState, TActivityState>;
}
interface ForkActivityState<TActivityState> {
    targetBranchName?: string;
    skipped?: true;
    activityState: TActivityState;
}

declare function createForkActivity<TStep extends BranchedStep, TGlobalState = object, TActivityState extends object = object>(stepType: TStep['type'], config: ForkActivityConfig<TStep, TGlobalState, TActivityState>): Activity<TGlobalState>;

type InterruptionActivityHandler<TStep extends Step, TGlobalState> = (step: TStep, globalState: TGlobalState) => Promise<void>;
interface InterruptionActivityConfig<TStep extends Step, TGlobalState> {
    handler: InterruptionActivityHandler<TStep, TGlobalState>;
}

declare function createInterruptionActivity<TStep extends Step = Step, TGlobalState = object>(stepType: TStep['type'], config: InterruptionActivityConfig<TStep, TGlobalState>): Activity<TGlobalState>;

type LoopActivityHandlerResult = boolean | InterruptResult;
type LoopActivityEventHandler<TStep extends SequentialStep, TGlobalState, TActivityState> = (step: TStep, globalState: TGlobalState, activityState: TActivityState) => Promise<void> | void;
type LoopActivityConditionHandler<TStep extends SequentialStep, TGlobalState, TActivityState> = (step: TStep, globalState: TGlobalState, activityState: TActivityState) => Promise<LoopActivityHandlerResult>;
interface LoopActivityConfig<TStep extends SequentialStep, TGlobalState, TActivityState extends object> {
    loopName: (step: TStep) => string;
    init: ActivityStateInitializer<TStep, TGlobalState, TActivityState>;
    condition: LoopActivityConditionHandler<TStep, TGlobalState, TActivityState>;
    onEnter?: LoopActivityEventHandler<TStep, TGlobalState, TActivityState>;
    onLeave?: LoopActivityEventHandler<TStep, TGlobalState, TActivityState>;
}
interface LoopActivityState<TActivityState> {
    continue?: boolean;
    activityState: TActivityState;
}

declare function createLoopActivity<TStep extends SequentialStep, TGlobalState = object, TActivityState extends object = object>(stepType: TStep['type'], config: LoopActivityConfig<TStep, TGlobalState, TActivityState>): Activity<TGlobalState>;

type ParallelActivityHandlerResult = InterruptResult | BranchNameResult[] | SkipResult;
type ParallelActivityHandler<TStep extends BranchedStep, TGlobalState, TActivityState> = (step: TStep, globalState: TGlobalState, activityState: TActivityState) => Promise<ParallelActivityHandlerResult>;
interface ParallelActivityState<TActivityState> {
    activityState: TActivityState;
    branchNames?: string[];
}
interface ParallelActivityConfig<TStep extends BranchedStep, TGlobalState, TActivityState extends object> {
    init: ActivityStateInitializer<TStep, TGlobalState, TActivityState>;
    handler: ParallelActivityHandler<TStep, TGlobalState, TActivityState>;
}

declare function createParallelActivity<TStep extends BranchedStep, TGlobalState = object, TActivityState extends object = object>(stepType: TStep['type'], config: ParallelActivityConfig<TStep, TGlobalState, TActivityState>): Activity<TGlobalState>;

declare class WorkflowMachineSnapshot<GlobalState> {
    readonly globalState: GlobalState;
    readonly unhandledError: MachineUnhandledError | undefined;
    private readonly stateValue;
    private statePaths?;
    constructor(globalState: GlobalState, unhandledError: MachineUnhandledError | undefined, stateValue: StateValue);
    /**
     * @returns The paths of all currently executing states.
     * @example `[ ['MAIN', 'STEP_1'], ['MAIN', 'STEP_2'] ]`
     */
    getStatePaths(): string[][];
    /**
     * @returns The path of the currently executing state.
     * If multiple states are currently executing, it returns `null`.
     * To get all paths of executing states, use the `getStatePaths` method.
     * @example `['MAIN', 'STEP_1']`
     * @example `null`
     */
    tryGetStatePath(): string[] | null;
    /**
     * @returns The ID of the currently executing state.
     * If multiple states are currently executing, it returns `null`.
     */
    tryGetCurrentStepId(): string | null;
    /**
     * @returns the list of ID of the currently executing steps.
     */
    getCurrentStepIds(): string[];
    /**
     * @returns `true` if the workflow machine is finished, otherwise `false`.
     */
    isFinished(): boolean;
    /**
     * @returns `true` if the workflow has failed, otherwise `false`.
     */
    isFailed(): boolean;
    /**
     * @returns `true` if the workflow is interrupted, otherwise `false`.
     */
    isInterrupted(): boolean;
}

declare class WorkflowMachineInterpreter<GlobalState> {
    private readonly interpreter;
    private readonly initState;
    constructor(interpreter: SequentialStateMachineInterpreter<GlobalState>, initState: State<MachineContext<GlobalState>> | undefined);
    start(): this;
    getSnapshot(): WorkflowMachineSnapshot<GlobalState>;
    serializeSnapshot(): SerializedWorkflowMachineSnapshot<GlobalState>;
    onDone(callback: () => void): this;
    onChange(callback: () => void): this;
    isRunning(): boolean;
    tryStop(): boolean;
    sendSignal<P extends SignalPayload>(event: string, payload: P): this;
    getNative(): SequentialStateMachineInterpreter<GlobalState>;
}

type BeforeSignalActivityHandler<TStep extends Step, GlobalState, ActivityState> = (step: TStep, globalState: GlobalState, activityState: ActivityState) => Promise<SignalActivityHandlerResult>;
type AfterSignalActivityHandler<TStep extends Step, GlobalState, ActivityState> = (step: TStep, globalState: GlobalState, activityState: ActivityState, signalPayload: SignalPayload) => Promise<SignalActivityHandlerResult>;
type SignalActivityHandlerResult = void | InterruptResult;
interface SignalActivityConfig<TStep extends Step, GlobalState, ActivityState extends object> {
    init: ActivityStateInitializer<TStep, GlobalState, ActivityState>;
    beforeSignal: BeforeSignalActivityHandler<TStep, GlobalState, ActivityState>;
    afterSignal: AfterSignalActivityHandler<TStep, GlobalState, ActivityState>;
}

declare function createSignalActivity<TStep extends Step, GlobalState = object, TActivityState extends object = object>(stepType: TStep['type'], config: SignalActivityConfig<TStep, GlobalState, TActivityState>): Activity<GlobalState>;
declare function signalSignalActivity<GlobalState, P extends SignalPayload>(interpreter: WorkflowMachineInterpreter<GlobalState>, payload: P): void;

declare class ActivityStateProvider<TStep extends Step, TGlobalState, TActivityState extends object> {
    private readonly step;
    private readonly init;
    constructor(step: TStep, init: ActivityStateInitializer<TStep, TGlobalState, TActivityState>);
    get(context: MachineContext<TGlobalState>, nodeId: string): TActivityState;
}

declare class ActivityNodeBuilderResolver<TGlobalState> {
    private readonly activitiesSet;
    private readonly sequenceNodeBuilder;
    private readonly builders;
    constructor(activitiesSet: ActivitySet<TGlobalState>, sequenceNodeBuilder: SequenceNodeBuilder<TGlobalState>);
    resolve(stepType: string): ActivityNodeBuilder<TGlobalState>;
}

declare function catchUnhandledError<TGlobalState>(callback: (context: MachineContext<TGlobalState>, event: EventObject, meta: InvokeMeta) => Promise<void>): (context: MachineContext<TGlobalState>, event: EventObject, meta: InvokeMeta) => Promise<void>;

declare const STATE_STEP_ID_PREFIX = "STEP_";
declare const STATE_BRANCH_ID_PREFIX = "BRANCH_";
declare function getStepNodeId(stepId: string): string;
declare function getBranchNodeId(stepId: string, branchName: string): string;

interface StartConfig<GlobalState> {
    init: GlobalStateInitializer<GlobalState>;
}
declare class WorkflowMachine<GlobalState> {
    private readonly definition;
    private readonly machine;
    constructor(definition: Definition, machine: SequentialStateMachine<GlobalState>);
    create(config: StartConfig<GlobalState>): WorkflowMachineInterpreter<GlobalState>;
    restore(context: MachineContext<GlobalState>): WorkflowMachineInterpreter<GlobalState>;
    deserializeSnapshot(serializedSnapshot: SerializedWorkflowMachineSnapshot<GlobalState>): WorkflowMachineInterpreter<GlobalState>;
    getNative(): SequentialStateMachine<GlobalState>;
}

interface BuildConfig {
    /**
     * @deprecated This property will be removed in the next minor version.
     */
    initialStepId?: string;
}
declare class WorkflowMachineBuilder<GlobalState> {
    private readonly sequenceNodeBuilder;
    constructor(sequenceNodeBuilder: SequenceNodeBuilder<GlobalState>);
    build(definition: Definition, config?: BuildConfig): WorkflowMachine<GlobalState>;
}
declare function createWorkflowMachineBuilder<GlobalState>(activitySet: ActivitySet<GlobalState>): WorkflowMachineBuilder<GlobalState>;

export { Activity, ActivityNodeBuilder, ActivityNodeBuilderFactory, ActivityNodeBuilderResolver, ActivityNodeConfig, ActivitySet, ActivityStateInitializer, ActivityStateProvider, AfterSignalActivityHandler, AtomActivityConfig, AtomActivityHandler, AtomActivityHandlerResult, BeforeSignalActivityHandler, BranchNameResult, BreakActivityConfig, BreakActivityHandler, BreakActivityHandlerResult, BreakActivityState, BreakResult, BuildConfig, BuildingContext, ContainerActivityConfig, ContainerActivityHandler, ContainerActivityHandlerResult, ForkActivityConfig, ForkActivityHandler, ForkActivityHandlerResult, ForkActivityState, GlobalStateInitializer, InterruptResult, InterruptionActivityConfig, InterruptionActivityHandler, LoopActivityConditionHandler, LoopActivityConfig, LoopActivityEventHandler, LoopActivityHandlerResult, LoopActivityState, MachineContext, MachineUnhandledError, ParallelActivityConfig, ParallelActivityHandler, ParallelActivityHandlerResult, ParallelActivityState, STATE_BRANCH_ID_PREFIX, STATE_FAILED_ID, STATE_FAILED_TARGET, STATE_FINISHED_ID, STATE_FINISHED_TARGET, STATE_INTERRUPTED_ID, STATE_INTERRUPTED_TARGET, STATE_STEP_ID_PREFIX, SequenceNodeBuilder, SequentialStateMachine, SequentialStateMachineInterpreter, SerializedWorkflowMachineSnapshot, SignalActivityConfig, SignalActivityHandlerResult, SignalPayload, SkipResult, StartConfig, WorkflowMachine, WorkflowMachineBuilder, WorkflowMachineInterpreter, branchName, break_, catchUnhandledError, createActivitySet, createAtomActivity, createAtomActivityFromHandler, createBreakActivity, createContainerActivity, createForkActivity, createInterruptionActivity, createLoopActivity, createParallelActivity, createSignalActivity, createWorkflowMachineBuilder, getBranchNodeId, getStepNodeId, interrupt, isBranchNameResult, isBreakResult, isInterruptResult, isSkipResult, signalSignalActivity, skip };
